// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "intention_receive.h"

#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_ball_and_player.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/player_param.h>

#include "neck_default_intercept_neck.h"
#include "neck_offensive_intercept_neck.h"

using namespace rcsc;

/*-------------------------------------------------------------------*/
/*!

 */
IntentionReceive::IntentionReceive( const Vector2D & target_point,
                                    const double & dash_power,
                                    const double & buf,
                                    const int max_step,
                                    const GameTime & start_time )
    : M_target_point( target_point )
    , M_dash_power( dash_power )
    , M_buffer( buf )
    , M_step( max_step )
    , M_last_execute_time( start_time )
{

}

/*-------------------------------------------------------------------*/
/*!

 */
bool
IntentionReceive::finished( const PlayerAgent * agent )
{
    if ( M_step <= 0 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": finished. time 0" );
        return true;
    }

    if ( agent->world().self().isKickable() )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": already kickable" );
        return true;
    }

    if ( agent->world().existKickableTeammate() )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": exist kickable teammate" );
        return true;
    }

    if ( agent->world().ball().distFromSelf() < 3.0 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": finished. ball very near" );
        return true;
    }

    if ( M_last_execute_time.cycle() < agent->world().time().cycle() - 1 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": finished. strange time." );
        return true;
    }

    if ( agent->world().self().pos().dist( M_target_point ) < M_buffer )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": finished. already there." );
        return true;
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
IntentionReceive::execute( PlayerAgent * agent )
{
    if ( M_step <= 0 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": execute. empty intention." );
        return false;
    }

    const WorldModel & wm = agent->world();

    M_step -= 1;
    M_last_execute_time = wm.time();

    agent->debugClient().setTarget( M_target_point );
    agent->debugClient().addMessage( "IntentionRecv" );

    dlog.addText( Logger::TEAM,
                  __FILE__": execute. try to receive" );

    int self_min = wm.interceptTable()->selfReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( self_min < 6 )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": execute. point very near. intercept" );
        agent->debugClient().addMessage( "IntentionRecv:Intercept" );
        Body_Intercept().execute( agent );

        if ( opp_min >= self_min + 3 )
        {
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        }
        else
        {
            agent->setNeckAction( new Neck_DefaultInterceptNeck
                                  ( new Neck_TurnToBallOrScan() ) );
        }

        return true;
    }

    if ( Body_Intercept().execute( agent ) )
    {
        dlog.addText( Logger::TEAM,
                      __FILE__": execute. intercept cycle=%d",
                      self_min );
        agent->debugClient().addMessage( "IntentionRecv%d:Intercept", M_step );
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        return true;
    }


    dlog.addText( Logger::TEAM,
                  __FILE__": execute. intercept cycle=%d. go to receive point",
                  self_min );
    agent->debugClient().addMessage( "IntentionRecv%d:GoTo", M_step );
    agent->debugClient().setTarget( M_target_point );
    agent->debugClient().addCircle( M_target_point, M_buffer );

    Body_GoToPoint( M_target_point,
                    M_buffer,
                    M_dash_power
                    ).execute( agent );
    agent->setNeckAction( new Neck_TurnToBallOrScan() );

    return true;
}




/*-------------------------------------------------------------------*/
/*!
 * 
 */
IntentionGetBall::IntentionGetBall( const Vector2D & target_point,
				    const double & dash_power,
				    const double & buf,
				    const int max_step,
				    const GameTime & start_time )
: M_target_point( target_point )
, M_dash_power( dash_power )
, M_buffer( buf )
, M_step( max_step )
, M_last_execute_time( start_time )
{
    
}

/*-------------------------------------------------------------------*/
/*!
 * 
 */

bool IntentionGetBall::finished(const rcsc::PlayerAgent* agent)
{
    if ( M_step <= 0 )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": finished. time 0" );
	return true;
    }
    
    if ( agent->world().self().isKickable() )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": already kickable" );
	return true;
    }
    
    if ( agent->world().existKickableTeammate() )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": exist kickable teammate" );
	return true;
    }
    
    if ( agent->world().ball().distFromSelf() < 1.0 )// )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": finished. ball very near" );
	return true;
    }
    
    if ( M_last_execute_time.cycle() < agent->world().time().cycle() - 1 )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": finished. strange time." );
	return true;
    }
    
    if ( agent->world().self().pos().dist( M_target_point ) < M_buffer )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": finished. already there." );
	return true;
    }
    
    return false;
}

/*-------------------------------------------------------------------*/
/*!
 * 
 */
bool
IntentionGetBall::execute( PlayerAgent * agent )
{
    if ( M_step <= 0 )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": execute. empty intention." );
	return false;
    }
    
    const WorldModel & wm = agent->world();
    
    M_step -= 1;
    M_last_execute_time = wm.time();
    
    agent->debugClient().setTarget( M_target_point );
    agent->debugClient().addMessage( "IntentionGB" );
    
    dlog.addText( Logger::TEAM,
		  __FILE__": execute. try to get ball" );
    
    int self_min = wm.interceptTable()->selfReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();
    
    if ( self_min < 6 )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": execute. point very near. intercept" );
	agent->debugClient().addMessage( "IntentionGetBall:Intercept" );
	Body_Intercept().execute( agent );
	
	if ( opp_min >= self_min + 3 )
	{
	    agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	}
	else
	{
	    agent->setNeckAction( new Neck_DefaultInterceptNeck
	    ( new Neck_TurnToBallOrScan() ) );
	}
	
	return true;
    }
    
    if ( Body_Intercept().execute( agent ) )
    {
	dlog.addText( Logger::TEAM,
		      __FILE__": execute. intercept cycle=%d",
		      self_min );
	agent->debugClient().addMessage( "IntentionGetBall%d:Intercept", M_step );
	agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	return true;
    }
    
    
    dlog.addText( Logger::TEAM,
		  __FILE__": execute. intercept cycle=%d. go to ball",
		  self_min );
    agent->debugClient().addMessage( "IntentionGB%d:GoTo", M_step );
    agent->debugClient().setTarget( M_target_point );
    agent->debugClient().addCircle( M_target_point, M_buffer );
    
    Body_GoToPoint( M_target_point,
		    M_buffer,
		    M_dash_power
    ).execute( agent );
    agent->setNeckAction( new Neck_TurnToBallOrScan() );
    
    return true;
}


/*-------------------------------------------------------------------*/


IntentionGoToPoint::IntentionGoToPoint( const Vector2D & target_point,
				    const double & dash_power,
				    const double & buf,
				    const int max_step,
				    const GameTime & start_time )
: M_target_point( target_point )
, M_dash_power( dash_power )
, M_buffer( buf )
, M_step( max_step )
, M_last_execute_time( start_time )
{

}

/*-------------------------------------------------------------------*/
/*!
 *
 */

bool IntentionGoToPoint::finished(const rcsc::PlayerAgent* agent)
{
	if ( M_step <= 0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. time 0" );
		return true;
	}

	if ( agent->world().self().pos().dist(M_target_point)<1)
	{
		dlog.addText( Logger::TEAM,
				__FILE__": already there" );
		return true;
	}

	if ( M_last_execute_time.cycle() < agent->world().time().cycle() - 1 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. strange time." );
		return true;
	}

	if ( agent->world().self().pos().dist( M_target_point ) < M_buffer )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. already there." );
		return true;
	}

	return false;
}

/*-------------------------------------------------------------------*/
/*!
 *
 */
bool
IntentionGoToPoint::execute( PlayerAgent * agent )
{
	if ( M_step <= 0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": execute. empty intention." );
		return false;
	}

	const WorldModel & wm = agent->world();

	M_step -= 1;
	M_last_execute_time = wm.time();

	agent->debugClient().setTarget( M_target_point );
	agent->debugClient().addMessage( "IntentionGTP" );

	dlog.addText( Logger::TEAM,
			__FILE__": execute. try to go to intended point" );

	Body_GoToPoint( M_target_point,M_buffer,M_dash_power).execute( agent );

	//	agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	agent->setNeckAction( new Neck_DefaultInterceptNeck
				( new Neck_TurnToBallOrScan(-1) ) );


	return true;

//
//	if ( Body_Intercept().execute( agent ) )
//	{
//		dlog.addText( Logger::TEAM,
//				__FILE__":IntentionGetBall execute. intercept cycle=%d",
//				M_step);//self_min );
//		agent->debugClient().addMessage( "IntentionGetBall%d:Intercept", M_step );
//		agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
//		return true;
//	}
//
//
//	dlog.addText( Logger::TEAM,
//			__FILE__": execute. intercept cycle=%d. go to ball",
//			self_min );
//	agent->debugClient().addMessage( "IntentionGB%d:GoTo", M_step );
//	agent->debugClient().setTarget( M_target_point );
//	agent->debugClient().addCircle( M_target_point, M_buffer );
//
//	Body_GoToPoint( M_target_point,M_buffer,M_dash_power).execute( agent );
//	agent->setNeckAction( new Neck_TurnToBallOrScan() );
//
//	return true;
}




/*-------------------------------------------------------------------*/


IntentionMark::IntentionMark( const PlayerObject * target_player,
				    const double & dash_power,
				    const double & buf,
				    const int max_step,
				    const GameTime & start_time )
: M_target_player( target_player )
, M_dash_power( dash_power )
, M_buffer( buf )
, M_step( max_step )
, M_last_execute_time( start_time )
{

}

/*-------------------------------------------------------------------*/
/*!
 *
 */

bool IntentionMark::finished(const rcsc::PlayerAgent* agent)
{
	if ( M_step <= 0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. time 0" );
		return true;
	}

//	if ( agent->world().self().pos().dist(M_target_point)<1)
//	{
//		dlog.addText( Logger::TEAM,
//				__FILE__": already there" );
//		return true;
//	}

	if ( M_last_execute_time.cycle() < agent->world().time().cycle() - 1 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. strange time." );
		return true;
	}

//	if ( agent->world().self().pos().dist( M_target_point ) < M_buffer )
//	{
//		dlog.addText( Logger::TEAM,
//				__FILE__": finished. already there." );
//		return true;
//	}

	return false;
}

/*-------------------------------------------------------------------*/
/*!
 *
 */
bool
IntentionMark::execute( PlayerAgent * agent )
{
	if ( M_step <= 0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": execute. empty intention." );
		return false;
	}

	const WorldModel & wm = agent->world();

	M_step -= 1;
	M_last_execute_time = wm.time();

	if(wm.self().isKickable())
	{
		if(Body_ClearBall().execute(agent))
			return true;
	}

	agent->debugClient().setTarget( M_target_player->pos());
	agent->debugClient().addMessage( "IntentionM" );

	dlog.addText( Logger::TEAM,
			__FILE__": execute. try to mark player %d ", M_target_player->unum());


	Vector2D marking_target=M_target_player->pos();
	Vector2D opp_pos=M_target_player->pos();

	Vector2D me_ball=wm.ball().pos()-wm.self().pos();

	//if(me_ball.norm()<=2)
	{
		Body_TackleToPoint tckl(opp_pos, 0.2,50);

		if(tckl.execute(agent))
		{
			dlog.addText( Logger::TEAM,__FILE__"tackling ball holder");
			return true;
		}
	}


	//compute marking point using M_target_player->pos() info
	Vector2D me_opp=opp_pos-wm.self().pos();

	Vector2D me_goal=ServerParam::i().ourTeamGoalPos()-wm.self().pos();

	Vector2D opp_ball=opp_pos-wm.ball().pos();
	Vector2D opp_goal=opp_pos-ServerParam::i().ourTeamGoalPos();

	double angle=0;
	double diff=0;

	bool opp_kickable=false;
	opp_kickable=(/*opp_goal.norm()<18  &&*/ M_target_player->isKickable() || (wm.getOpponentNearestToBall(100,true) && wm.getOpponentNearestToBall(100,true)->unum()==M_target_player->unum()));

	if(!opp_kickable)
		diff=fabs(me_ball.dir().degree()-me_opp.dir().degree());
	else
		diff=fabs(me_goal.dir().degree()-me_opp.dir().degree());


	if(diff>180)
	{
		diff=360-diff;
	}

	dlog.addText( Logger::TEAM,__FILE__" marking -- me_b %.1f me_o %.1f diff/? %.1f - kickable? %d",me_ball.dir().degree(),me_opp.dir().degree(),diff/(opp_kickable?2:3),opp_kickable?1:0);

	if (
			(!opp_kickable &&
					(
							(me_ball.dir().degree()>me_opp.dir().degree() && !(me_ball.dir().degree()>0 && me_opp.dir().degree()<0 && fabs(me_ball.dir().degree()-me_opp.dir().degree())>180))
							||
							(me_ball.dir().degree()<0 && me_opp.dir().degree()>0 && fabs(me_ball.dir().degree()-me_opp.dir().degree())>180)
					)
			)
			||
			(opp_kickable &&
					(
							(me_goal.dir().degree()>me_opp.dir().degree() && !(me_goal.dir().degree()>0 && me_opp.dir().degree()<0 && fabs(me_goal.dir().degree()-me_opp.dir().degree())>180))
							||
							(me_goal.dir().degree()<0 && me_opp.dir().degree()>0 && fabs(me_goal.dir().degree()-me_opp.dir().degree())>180)
					)
			)
	)
		//greater or b is smaller but still need to add
		angle=me_opp.dir().degree()+diff/(opp_kickable?2:3);
	else
		angle=me_opp.dir().degree()-diff/(opp_kickable?2:3);


	double r=0;
	if(!opp_kickable)
		r=me_ball.norm()*me_opp.norm()*(1-(pow(opp_ball.norm(),2)/std::pow((me_ball.norm()+me_opp.norm()),2)));
	//r=me_ball.norm()-opp_ball.norm()+(M_target_player->unum()!=-1?0.33:0.37)*opp_ball.norm();
	else
		r=me_goal.norm()*me_opp.norm()*(1-(pow(opp_goal.norm(),2)/std::pow((me_opp.norm()+me_goal.norm()),2)));
	//r=me_goal.norm()-opp_goal.norm()+0.33*opp_goal.norm();

	if(r<0)r=0;
//	{
//		angle+=180;
//		if(angle>180)angle-=360;
//		r*=-1;
//	}



	marking_target.setPolar(r,AngleDeg(angle));

	marking_target=marking_target+wm.self().pos();
//	if(opp_pos.dist(ServerParam::i().ourTeamGoalPos())>15)
//		marking_target.x=marking_target.x-2;

	Vector2D cover=opp_goal;
	cover.setLength(2);


	//if the opponent is not in shooting range, keep behind him
	if(opp_pos.dist(ServerParam::i().ourTeamGoalPos())>15)
		//dribble_target.x=dribble_target.x-2;
		marking_target=marking_target+cover;


	if(wm.ball().pos().x>0)	//the ball is in their half
	{
		//tell defenders not to advance much
		if(wm.self().unum()<=4)
		{
			if(marking_target.x>-15)
				marking_target.x=std::min(-15,(int)marking_target.x-7);
		}
	}
	dlog.addText( Logger::TEAM,__FILE__" marking -- marking target (%.1f,%.1f) -- r (%.1f) cover (%.1f,%.1f)",marking_target.x,marking_target.y,r,cover.x,cover.y);


	//marking_target=marking_target+wm.self().pos();

	Body_GoToPoint( marking_target,
			M_buffer,
			M_dash_power,
			opp_ball.norm()<5?ServerParam::i().DEFAULT_PLAYER_SPEED_MAX+PlayerParam::i().playerSpeedMaxDeltaMax():-1).execute( agent );

	//	agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	agent->setNeckAction( new Neck_DefaultInterceptNeck( new Neck_TurnToBallAndPlayer(M_target_player) ));


	return true;

	//
	//	if ( Body_Intercept().execute( agent ) )
	//	{
	//		dlog.addText( Logger::TEAM,
	//				__FILE__":IntentionGetBall execute. intercept cycle=%d",
	//				M_step);//self_min );
	//		agent->debugClient().addMessage( "IntentionGetBall%d:Intercept", M_step );
	//		agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
	//		return true;
	//	}
	//
	//
	//	dlog.addText( Logger::TEAM,
	//			__FILE__": execute. intercept cycle=%d. go to ball",
	//			self_min );
	//	agent->debugClient().addMessage( "IntentionGB%d:GoTo", M_step );
	//	agent->debugClient().setTarget( M_target_point );
	//	agent->debugClient().addCircle( M_target_point, M_buffer );
	//
	//	Body_GoToPoint( M_target_point,M_buffer,M_dash_power).execute( agent );
	//	agent->setNeckAction( new Neck_TurnToBallOrScan() );
	//
	//	return true;
}




/*-------------------------------------------------------------------*/


IntentionBlock::IntentionBlock( const int max_step,
				    const GameTime & start_time )
:M_step( max_step )
, M_last_execute_time( start_time )
{

}

/*-------------------------------------------------------------------*/
/*!
 *
 */

bool IntentionBlock::finished(const rcsc::PlayerAgent* agent)
{
	if ( M_step <= 0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. time 0" );
		return true;
	}

//	if ( agent->world().self().pos().dist(M_target_point)<1)
//	{
//		dlog.addText( Logger::TEAM,
//				__FILE__": already there" );
//		return true;
//	}

	if ( M_last_execute_time.cycle() < agent->world().time().cycle() - 1 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": finished. strange time." );
		return true;
	}

//	if ( agent->world().self().pos().dist( M_target_point ) < M_buffer )
//	{
//		dlog.addText( Logger::TEAM,
//				__FILE__": finished. already there." );
//		return true;
//	}

	return false;
}

/*-------------------------------------------------------------------*/
/*!
 *
 */
bool
IntentionBlock::execute( PlayerAgent * agent )
{
	if ( M_step <= 0 )
	{
		dlog.addText( Logger::TEAM,
				__FILE__": execute. empty intention." );
		return false;
	}

	const WorldModel & wm = agent->world();

	M_step -= 1;
	M_last_execute_time = wm.time();

	if(wm.self().isKickable())
	{
		if(Body_ClearBall().execute(agent))
			return true;
	}


	Vector2D blocking_target=wm.self().pos();

	dlog.addText( Logger::TEAM,
			__FILE__": execute. try to block");

	Vector2D goal_c=ServerParam::instance().ourTeamGoalPos();
	Vector2D me_goal=goal_c-wm.self().pos();

	Vector2D opp_pos=wm.ball().pos();

	if(wm.getOpponentNearestToBall(100,true))
	{
		opp_pos=opp_pos+(wm.getOpponentNearestToBall(100,true))->pos();
		opp_pos.scale(0.5);	//the point between the player and the ball
	}
	double dist=0;
	const rcsc::PlayerObject * mate=wm.getTeammateNearestTo(wm.getOpponentNearestToBall(100,true),100,&dist);
	if(dist>=wm.self().pos().dist(opp_pos))	//I'm closer -- block
	{
		Vector2D me_opp=opp_pos-wm.self().pos();

		double angle=0;
		double diff=fabs(me_goal.dir().degree()-me_opp.dir().degree());


		dlog.addText( Logger::TEAM,__FILE__" blocking -- me_goal %.1f me_opp %.1f diff %.1f",me_goal.dir().degree(),me_opp.dir().degree(),diff);


		if(diff>180)
		{
			diff=360-diff;
			//the angles have different signs
			if (me_goal.dir().degree()>me_opp.dir().degree()) 	//positive
			{
				if(fabs(me_goal.dir().degree())<fabs(me_opp.dir().degree())) //smaller
					angle=me_goal.dir().degree()+diff/2;
				else
					angle=me_opp.dir().degree()-diff/2;
			}
			else
			{
				if(fabs(me_goal.dir().degree())<fabs(me_opp.dir().degree())) //smaller
					angle=me_goal.dir().degree()-diff/2;
				else
					angle=me_opp.dir().degree()+diff/2;
			}
		}
		else
		{
			if(me_goal.dir().degree()*me_opp.dir().degree() >0) 	//of same sign
			{
				if(me_goal.dir().degree()<me_opp.dir().degree()) //smaller
					angle=me_goal.dir().degree()+diff/2;
				else
					angle=me_opp.dir().degree()+diff/2;
			}
			else	//different signs
			{
				if(me_goal.dir().degree()<me_opp.dir().degree()) //negative
					angle=me_goal.dir().degree()+diff/2;
				else
					angle=me_opp.dir().degree()+diff/2;
			}

		}

		dlog.addText( Logger::TEAM,__FILE__" blocking target -- angle %.1f",angle);
		Vector2D opp_goal=ServerParam::i().ourTeamGoalPos()-opp_pos;

		double r=0;
		//r=me_goal.norm()-opp_goal.norm()+0.33*opp_goal.norm();
		r=me_goal.norm()*me_opp.norm()*(1-(pow(opp_goal.norm(),2)/std::pow((me_opp.norm()+me_goal.norm()),2)));

		if(r<0)r=0;

		blocking_target.setPolar(r,AngleDeg(angle));
		blocking_target=blocking_target+wm.self().pos();
		dlog.addText( Logger::TEAM,__FILE__" blocking target=(%.1f %.1f)",
				blocking_target.x, blocking_target.y );

	}



	dlog.addText( Logger::TEAM,
			__FILE__" (PredicateComplex Bhv_ChainAction) final moving target=(%.1f %.1f)",
			blocking_target.x, blocking_target.y );

	if(wm.ball().pos().x>0)	//the ball is in their half
			{
				//tell defenders not to advance much
				if(wm.self().unum()<=4)
				{
					if(blocking_target.x>-15)
						blocking_target.x=std::min(-15,(int)blocking_target.x-7);
				}
			}
	//dribble needs ball to be kickable

	//
	//    				Dribble d(wm.self().unum(),move_target,0.15*wm.self().pos().dist(move_target),1,1,10);
	//    				if ( Bhv_NormalDribble(d).execute( agent ) )
	//    				{
	//    					return true;
	//    				}
	//

	Vector2D me_targ=blocking_target-wm.self().pos();


	if(blocking_target.x>ServerParam::i().DEFAULT_PITCH_LENGTH/2)
		blocking_target.x=ServerParam::i().DEFAULT_PITCH_LENGTH/2-0.5;
	else if(blocking_target.x<-1*ServerParam::i().DEFAULT_PITCH_LENGTH/2)
		blocking_target.x=-1*ServerParam::i().DEFAULT_PITCH_LENGTH/2+0.5;

	if(blocking_target.y>ServerParam::i().DEFAULT_PITCH_WIDTH/2)
		blocking_target.y=ServerParam::i().DEFAULT_PITCH_WIDTH/2-0.5;
	else if(blocking_target.y<-1*ServerParam::i().DEFAULT_PITCH_WIDTH/2)
		blocking_target.y=-1*ServerParam::i().DEFAULT_PITCH_WIDTH/2+0.5;



	double dist_thr = wm.ball().distFromSelf() * 0.1;
	if ( dist_thr < 1.0 ) dist_thr = 1.0;
	if(Body_GoToPoint(blocking_target,dist_thr, ServerParam::i().MAX_DASH_POWER).execute(agent))
		//if(agent->doDash(75,me_targ.dir()-wm.self().body()))
	{
		dlog.addText( Logger::TEAM,__FILE__" dashing to dribble target=(%.1f %.1f) using angle %f -   %f ",
				blocking_target.x, blocking_target.y,me_targ.dir().degree(),wm.self().body().degree());
		return true;
	}
	//		else
	//			{
	//			dlog.addText( Logger::TEAM,__FILE__" goToPoint failed");
	//	        Body_TurnToBall().execute( agent );
	//
	//			}

	if ( wm.existKickableOpponent() && wm.ball().distFromSelf() < 18.0 )
	{
		agent->setNeckAction( new Neck_TurnToBall() );
	}
	else
	{
		agent->setNeckAction( new Neck_TurnToBallOrScan() );
	}
}
