// -*-c++-*-

/*!
  \file sample_communication.h
  \brief sample communication planner Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef SAMPLE_COMMUNICATION_H
#define SAMPLE_COMMUNICATION_H

#include "communication.h"
#include "rcsc/player/say_message_builder.h"


#include <rcsc/game_time.h>
#include <rcsc/types.h>

namespace rcsc {
class WorldModel;
}

class SampleCommunication
    : public Communication {
private:

    int M_current_sender_unum;
    int M_next_sender_unum;
    rcsc::GameTime M_ball_send_time;
    rcsc::GameTime M_teammate_send_time[12];
    rcsc::GameTime M_opponent_send_time[12];

public:

    SampleCommunication();
    ~SampleCommunication();

    virtual
    bool execute( rcsc::PlayerAgent * agent );

    int currentSenderUnum() const { return M_current_sender_unum; }

    int nextSenderUnum() const { return M_next_sender_unum; }

private:
    void updateCurrentSender( const rcsc::PlayerAgent * agent );

    void updatePlayerSendTime( const rcsc::WorldModel & wm,
                               const rcsc::SideID side,
                               const int unum );

    bool shouldSayBall( const rcsc::PlayerAgent * agent );
    bool shouldSayOpponentGoalie( const rcsc::PlayerAgent * agent );
    bool goalieSaySituation( const rcsc::PlayerAgent * agent );

    bool sayBallAndPlayers( rcsc::PlayerAgent * agent );
    bool sayBall( rcsc::PlayerAgent * agent );
    bool sayPlayers( rcsc::PlayerAgent * agent );
    bool sayGoalie( rcsc::PlayerAgent * agent );

    bool sayIntercept( rcsc::PlayerAgent * agent );
    bool sayOffsideLine( rcsc::PlayerAgent * agent );
    bool sayDefenseLine( rcsc::PlayerAgent * agent );

    bool sayOpponents( rcsc::PlayerAgent * agent );
    bool sayTwoOpponents( rcsc::PlayerAgent * agent );
    bool sayThreeOpponents( rcsc::PlayerAgent * agent );

    bool saySelf( rcsc::PlayerAgent * agent );

    bool sayStamina( rcsc::PlayerAgent * agent );
    bool sayRecovery( rcsc::PlayerAgent * agent );
    
    bool sayValue( rcsc::PlayerAgent* agent, int index, double value );

    void attentiontoSomeone( rcsc::PlayerAgent * agent );
};


/*-----------------------------------------------------------------------------*/

#include "rcsc/common/audio_memory.h"
#include "rcsc/common/audio_codec.h"
//class ValueMessageParser
/*!
 *   format:
 *  "v<value:2>"
 *  the length of message == 3
 */
class ValueMessageParser
: public rcsc::SayMessageParser {
private:
    
    //! pointer to the audio memory
    boost::shared_ptr< rcsc::AudioMemory > M_memory;
    
public:
    
    /*!
     *      \brief construct with audio memory
     *      \param memory pointer to the memory
     */
    explicit
    ValueMessageParser( boost::shared_ptr< rcsc::AudioMemory > memory );
    
    /*!
     *      \brief get the header character.
     *      \return header character.
     */
    static
    char vheader() { return 'v'; }
    
    /*!
     *      \brief get the header character.
     *      \return header character.
     */
    char header() const { return vheader(); }
    
    /*!
     *      \brief get the length of this message.
     *      \return the length of encoded message
     */
    static
    int vlength() { return 3; }
    
    /*!
     *      \brief virtual method which analyzes audio messages.
     *      \param sender sender's uniform number
     *      \param dir sender's direction
     *      \param msg raw audio message
     *      \param current current game time
     *      \retval bytes read if success
     *      \retval 0 message ID is not match. other parser should be tried.
     *      \retval -1 failed to parse
     */
    int parse( const int sender,
	       const double & dir,
	       const char * msg,
	       const rcsc::GameTime & current );
    
};

/*----------------------------------------------------------------------------------------------------*/
//class ValueMessage
/*!
  format:
  "v<value:2>"
  the length of message == 3
*/
class ValueMessage : public rcsc::SayMessage 
{
private:
    int M_id;	//id of Tstate
    double M_value; //value of Tstate

public:

    explicit
    ValueMessage( const int index, const double & value )
        : M_value( value ),
	  M_id(index)
      { }

   
    char header() const
      {
          return ValueMessageParser::vheader();
      }

    /*!
      \brief get the length of this message.
      \return the length of encoded message
    */
    static
    int vlength()
      {
          return ValueMessageParser::vlength();
      }

    /*!
      \brief get the length of this message
      \return the length of encoded message
    */
    int length() const
      {
          return vlength();
      }

    /*!
      \brief append this info to the audio message
      \param to reference to the message string instance
      \return result status of encoding
    */
    bool toStr( std::string & to ) const;

    /*!
      \brief append the debug message
      \param os reference to the output stream
      \return reference to the output stream
     */
    std::ostream & printDebug( std::ostream & os ) const;

};




#include "rcsc/common/audio_memory.h"
#include "rcsc/common/audio_codec.h"
//class MarkingMessageParser
/*!
 *   format:
 *  "m<unum:2>"
 *  the length of message == 3
 */
class MarkingMessageParser
: public rcsc::SayMessageParser {
private:

    //! pointer to the audio memory
    boost::shared_ptr< rcsc::AudioMemory > M_memory;

public:

    /*!
     *      \brief construct with audio memory
     *      \param memory pointer to the memory
     */
    explicit
    MarkingMessageParser( boost::shared_ptr< rcsc::AudioMemory > memory );

    /*!
     *      \brief get the header character.
     *      \return header character.
     */
    static
    char mheader() { return 'm'; }

    /*!
     *      \brief get the header character.
     *      \return header character.
     */
    char header() const { return mheader(); }

    /*!
     *      \brief get the length of this message.
     *      \return the length of encoded message
     */
    static
    int mlength() { return 3; }

    bool valid();
    /*!
     *      \brief virtual method which analyzes audio messages.
     *      \param sender sender's uniform number
     *      \param dir sender's direction
     *      \param msg raw audio message
     *      \param current current game time
     *      \retval bytes read if success
     *      \retval 0 message ID is not match. other parser should be tried.
     *      \retval -1 failed to parse
     */
    int parse( const int sender,
	       const double & dir,
	       const char * msg,
	       const rcsc::GameTime & current );

};


/*----------------------------------------------------------------------------------------------------*/
//class MarkingMessage
/*!
  format:
  "m<value:2>"
  the length of message == 3
*/
class MarkingMessage : public rcsc::SayMessage
{
private:

    int M_marked_unum; //unum of marked opponent
    //int M_id;	//id of Tstate

public:

    explicit
    MarkingMessage( const int unum)
        : M_marked_unum( unum )
      {if (unum<1 || unum >11) M_marked_unum=-1;}

    explicit
    MarkingMessage()
    : M_marked_unum( -1 )
    { }

    char header() const
      {
          return MarkingMessageParser::mheader();
      }

    bool valid()
    {
    	return (!(M_marked_unum==-1));
    }

    /*!
      \brief get the length of this message.
      \return the length of encoded message
    */
    static
    int mlength()
      {
          return MarkingMessageParser::mlength();
      }

    /*!
      \brief get the length of this message
      \return the length of encoded message
    */
    int length() const
      {
          return mlength();
      }

    /*!
      \brief append this info to the audio message
      \param to reference to the message string instance
      \return result status of encoding
    */
    bool toStr( std::string & to ) const;

    /*!
      \brief append the debug message
      \param os reference to the output stream
      \return reference to the output stream
     */
    std::ostream & printDebug( std::ostream & os ) const;

};


#endif
